#ifndef CYGONCE_DEVS_FLASH__W29EEXXX_INL
#define CYGONCE_DEVS_FLASH__W29EEXXX_INL
//==========================================================================
//
//      flash_winbond_w29eexxx.inl
//
//      Winbond W29EExxx series flash driver
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    jfoster
// Contributors: gthomas, jskov
// Date:         2002-10-21
// Purpose:      
// Description:  Winbond W29EExxx flash driver.
//              
// Notes:        FLASH_P2V is not properly used.
//               Needs locking.
// NOTE: For this driver, the flash block sizes are very small.  If using
//       RedBoot, configure the CYGNUM_REDBOOT_FLASH_CONFIG_BLOCK parameter
//       to a value that will keep the RedBoot configuration from overwriting
//       the FIS directory (-17 for example).
//       Also configure the CYGNUM_REDBOOT_FIS_DIRECTORY_ENTRY_SIZE
//       parameter to a value that will permit more than one FIS directory
//       entry (44 is the minimum).
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/hal.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_cache.h>
#include CYGHWR_MEMORY_LAYOUT_H
#include <cyg/infra/cyg_ass.h>
#include <cyg/hal/hal_intr.h>

#define  _FLASH_PRIVATE_
#include <cyg/io/flash.h>


//----------------------------------------------------------------------------
// Common device details.
#define FLASH_Read_ID                   FLASHWORD( 0x60 )
#define FLASH_Read_ID_Exit              FLASHWORD( 0xF0 )
#define FLASH_Reset                     FLASHWORD( 0xFF )
#define FLASH_Program                   FLASHWORD( 0xA0 )
#define FLASH_Block_Erase               FLASHWORD( 0xA0 )
#define FLASH_Chip_Erase                FLASHWORD( 0x10 )
#define FLASH_Softare_Protect_Enable    FLASHWORD( 0x20 )

#define FLASH_Data                      FLASHWORD( 0x80 ) // Data complement
#define FLASH_Busy                      FLASHWORD( 0x40 ) // "Toggle" bit

#define FLASH_Setup_Addr1               (0x5555)
#define FLASH_Setup_Addr2               (0x2AAA)
#define FLASH_Setup_Code1               FLASHWORD( 0xAA )
#define FLASH_Setup_Code2               FLASHWORD( 0x55 )
#define FLASH_Setup_Erase               FLASHWORD( 0x80 )

// Platform code must define the below
// #define CYGNUM_FLASH_INTERLEAVE      : Number of interleaved devices (in parallel)
// #define CYGNUM_FLASH_SERIES          : Number of devices in series
// #define CYGNUM_FLASH_BASE            : Address of first device
// And select one of the below device variants

#ifdef CYGPKG_DEVS_FLASH_WINBOND_W29EE512
# define FLASH_DEV_BLOCK_SIZE           (128)
# define FLASH_BLOCK_SIZE               (FLASH_DEV_BLOCK_SIZE*CYGNUM_FLASH_INTERLEAVE)
# define FLASH_NUM_REGIONS              (512)
# define CYGNUM_FLASH_WIDTH             (8)
# define CYGNUM_FLASH_BLANK             (1)
# define CYGNUM_FLASH_ID_MANUFACTURER   FLASHWORD(0xDA)
# define CYGNUM_FLASH_ID_DEVICE         FLASHWORD(0xC8)
#endif

#ifdef CYGPKG_DEVS_FLASH_WINBOND_W29EE011
# define FLASH_DEV_BLOCK_SIZE           (128)
# define FLASH_BLOCK_SIZE               (FLASH_DEV_BLOCK_SIZE*CYGNUM_FLASH_INTERLEAVE)
# define FLASH_NUM_REGIONS              (1024)
# define CYGNUM_FLASH_WIDTH             (8)
# define CYGNUM_FLASH_BLANK             (1)
# define CYGNUM_FLASH_ID_MANUFACTURER   FLASHWORD(0xDA)
# define CYGNUM_FLASH_ID_DEVICE         FLASHWORD(0xC1)
#endif

#define FLASH_DEVICE_SIZE               (FLASH_BLOCK_SIZE*FLASH_NUM_REGIONS)
#define CYGNUM_FLASH_BASE_MASK          (~(FLASH_DEVICE_SIZE-1))
#define CYGNUM_FLASH_DEVICES            (CYGNUM_FLASH_INTERLEAVE*CYGNUM_FLASH_SERIES)

//----------------------------------------------------------------------------
// Now that device properties are defined, include magic for defining
// accessor type and constants.
#include <cyg/io/flash_dev.h>

//----------------------------------------------------------------------------
// Private buffer for writing a partial flash device block.
static flash_data_t flash_xfer_buf[FLASH_DEV_BLOCK_SIZE];

//----------------------------------------------------------------------------
// Functions that put the flash device into non-read mode must reside
// in RAM.
void flash_query(void* data) __attribute__ ((section (".2ram.flash_query")));
int  flash_erase_block(void* block, unsigned int size) 
    __attribute__ ((section (".2ram.flash_erase_block")));
int  flash_program_buf(void* addr, void* data, int len)
    __attribute__ ((section (".2ram.flash_program_buf")));


//----------------------------------------------------------------------------
// Initialize driver details
int
flash_hwr_init(void)
{
    flash_data_t id[2];

    flash_dev_query(id);

    // Check that flash_id data is matching the one the driver was
    // configured for.
    if (id[0] != CYGNUM_FLASH_ID_MANUFACTURER
        || id[1] != CYGNUM_FLASH_ID_DEVICE)
        return FLASH_ERR_DRV_WRONG_PART;

    // Hard wired for now
    flash_info.block_size = FLASH_BLOCK_SIZE;
    flash_info.blocks = FLASH_NUM_REGIONS;
    flash_info.start = (void *)CYGNUM_FLASH_BASE;
    flash_info.end = (void *)(CYGNUM_FLASH_BASE+ (FLASH_NUM_REGIONS * FLASH_BLOCK_SIZE * CYGNUM_FLASH_SERIES));
    return FLASH_ERR_OK;
}

//----------------------------------------------------------------------------
// Map a hardware status to a package error
int
flash_hwr_map_error(int e)
{
    return e;
}


//----------------------------------------------------------------------------
// See if a range of FLASH addresses overlaps currently running code
bool
flash_code_overlaps(void *start, void *end)
{
    extern unsigned char _stext[], _etext[];

    return ((((CYG_ADDRESS)&_stext >= (CYG_ADDRESS)start) &&
             ((CYG_ADDRESS)&_stext < (CYG_ADDRESS)end)) ||
            (((CYG_ADDRESS)&_etext >= (CYG_ADDRESS)start) &&
             ((CYG_ADDRESS)&_etext < (CYG_ADDRESS)end)));
}

//----------------------------------------------------------------------------
// Flash Query
//
// Only reads the manufacturer and part number codes for the first
// device(s) in series. It is assumed that any devices in series
// will be of the same type.

void
flash_query(void* data)
{
    volatile flash_data_t *ROM;
    flash_data_t* id = (flash_data_t*) data;
    int i;
#if	defined(CYG_HAL_STARTUP_ROM)
    cyg_uint32 pri;
#endif

    ROM = (volatile flash_data_t*) CYGNUM_FLASH_BASE;

#if	defined(CYG_HAL_STARTUP_ROM)
    HAL_DISABLE_INTERRUPTS(pri);
#endif
    ROM[FLASH_Setup_Addr1] = FLASH_Setup_Code1;
    ROM[FLASH_Setup_Addr2] = FLASH_Setup_Code2;
    ROM[FLASH_Setup_Addr1] = FLASH_Setup_Erase;
    ROM[FLASH_Setup_Addr1] = FLASH_Setup_Code1;
    ROM[FLASH_Setup_Addr2] = FLASH_Setup_Code2;
    ROM[FLASH_Setup_Addr1] = FLASH_Read_ID;

    // FIXME: 10us delay
    for (i = 10000; i > 0; i--);

    // Manufacturers' code
    id[0] = ROM[0];
    // Part number
    id[1] = ROM[1];

    ROM[FLASH_Setup_Addr1] = FLASH_Setup_Code1;
    ROM[FLASH_Setup_Addr2] = FLASH_Setup_Code2;
    ROM[FLASH_Setup_Addr1] = FLASH_Read_ID_Exit;
#if	defined(CYG_HAL_STARTUP_ROM)
    HAL_RESTORE_INTERRUPTS(pri);
#endif

    // FIXME: 10us delay
    for (i = 10000; i > 0; i--);
}

//----------------------------------------------------------------------------
// Erase Block
int
flash_erase_block(void* block, unsigned int len)
{
    volatile flash_data_t* ROM;
    volatile flash_data_t* addr_ptr = (volatile flash_data_t*) block;
    volatile flash_data_t* addr_ptr2;
#if	defined(CYG_HAL_STARTUP_ROM)
    cyg_uint32 pri;
#endif

    int res = FLASH_ERR_OK;

    // Base address of device(s) being programmed. 
    ROM = (volatile flash_data_t*)((CYG_ADDRESS)block & ~(FLASH_DEVICE_SIZE-1));

    while ((FLASH_ERR_OK == res) && (len > 0)) {
        int len2, i, timeout;
        flash_data_t state, prev_state;

        addr_ptr2 = addr_ptr;
        len2 = len;

#if	defined(CYG_HAL_STARTUP_ROM)
        HAL_DISABLE_INTERRUPTS(pri);
#endif
        // Program data [byte] - 4 step sequence
        ROM[FLASH_Setup_Addr1] = FLASH_Setup_Code1;
        ROM[FLASH_Setup_Addr2] = FLASH_Setup_Code2;
        ROM[FLASH_Setup_Addr1] = FLASH_Block_Erase;

        // Always load FLASH_DEV_BLOCK_SIZE FLASHWORDs
        for (i = 0; i < FLASH_DEV_BLOCK_SIZE; i++) {
            *addr_ptr2++ = FLASH_BlankValue;
        }

        // Wait for 300us for programming to begin.
        // FIXME: 300us delay
        for (i = 10000; i > 0; i--);

        // Wait for completion (bit 6 stops toggling)
        timeout = 5000000;
        prev_state = *addr_ptr & FLASH_Busy;
        while (1) {
            state = *addr_ptr & FLASH_Busy;
            if (prev_state == state) {
                break;
            }
            if (--timeout == 0) {
                res = FLASH_ERR_DRV_TIMEOUT;
                break;
            }
            prev_state = state;
        }
#if	defined(CYG_HAL_STARTUP_ROM)
        HAL_RESTORE_INTERRUPTS(pri);
#endif

        // Verify loaded data bytes
        for (i = 0; (len > 0) && (i < FLASH_DEV_BLOCK_SIZE) ; i++) {
            if (*addr_ptr != FLASH_BlankValue) {
                // Only update return value if erase operation was OK
                if (FLASH_ERR_OK == res) res = FLASH_ERR_DRV_VERIFY;
                break;
            }
            addr_ptr++;
            len -= sizeof(*addr_ptr);
        }
    }

    return res;
}

//----------------------------------------------------------------------------
// Program Buffer
int
flash_program_buf(void* addr, void* data, int len)
{
    volatile flash_data_t* ROM;
    volatile flash_data_t* addr_ptr = (volatile flash_data_t*) addr;
    volatile flash_data_t* data_ptr = (volatile flash_data_t*) data;
    volatile flash_data_t* addr_ptr2;
    volatile flash_data_t* data_ptr2;
    int bindex;
    int nbytes;
#if	defined(CYG_HAL_STARTUP_ROM)
    cyg_uint32 pri;
#endif
    int res = FLASH_ERR_OK;

    // Base address of device(s) being programmed. 
    ROM = (volatile flash_data_t*)((CYG_ADDRESS)addr & ~(FLASH_DEVICE_SIZE-1));

    while ((FLASH_ERR_OK == res) && (len > 0)) {
        int len2, i, timeout;
        flash_data_t state, prev_state;

        // Copy the current contents of the flash block into the xfer buffer.
        memcpy((void *)flash_xfer_buf, (void *)((CYG_ADDRESS)addr_ptr & ~(FLASH_BLOCK_SIZE-1)), sizeof(flash_xfer_buf));
        // Copy the new data over the old data in the transfer buffer.
        bindex = (int)((CYG_ADDRESS)addr_ptr - ((CYG_ADDRESS)addr_ptr & ~(FLASH_BLOCK_SIZE-1)));
        CYG_ASSERTC((bindex<FLASH_BLOCK_SIZE)&&(bindex>=0));
        CYG_ASSERTC(FLASH_BLOCK_SIZE==sizeof(flash_xfer_buf));
        nbytes = sizeof(flash_xfer_buf) - bindex;
        if (nbytes > len) nbytes = len;
        memcpy((void *)((cyg_uint8 *)flash_xfer_buf + bindex), (void *)data_ptr, nbytes);
        // Now write the transfer buffer to this block of flash.

        addr_ptr2 = (volatile flash_data_t*)((CYG_ADDRESS)addr_ptr & ~(FLASH_BLOCK_SIZE-1));
        addr_ptr = addr_ptr2;			// Align to flash block start.
        data_ptr2 = flash_xfer_buf;	// Data comes from the transfer buffer
        len2 = len;

#if	defined(CYG_HAL_STARTUP_ROM)
        HAL_DISABLE_INTERRUPTS(pri);
#endif
        // Program data [byte] - 4 step sequence
        ROM[FLASH_Setup_Addr1] = FLASH_Setup_Code1;
        ROM[FLASH_Setup_Addr2] = FLASH_Setup_Code2;
        ROM[FLASH_Setup_Addr1] = FLASH_Program;

        // Always load FLASH_DEV_BLOCK_SIZE FLASHWORDs
        for (i = 0; i < FLASH_DEV_BLOCK_SIZE; i++) {
            if (len2 > 0) {
                *addr_ptr2++ = *data_ptr2++;
                len2 -= sizeof(*data_ptr2);
            } else {
                *addr_ptr2++ = FLASH_BlankValue;
            }
        }

        // Wait for 300us for programming to begin.
        // FIXME: 300us delay
        for (i = 10000; i > 0; i--);

        // Wait for completion (bit 6 stops toggling)
        timeout = 5000000;
        prev_state = *addr_ptr & FLASH_Busy;
        while (1) {
            state = *addr_ptr & FLASH_Busy;
            if (prev_state == state) {
                break;
            }
            if (--timeout == 0) {
                res = FLASH_ERR_DRV_TIMEOUT;
                break;
            }
            prev_state = state;
        }
#if	defined(CYG_HAL_STARTUP_ROM)
        HAL_RESTORE_INTERRUPTS(pri);
#endif

        // Verify loaded data bytes
        for (i = 0; (len > 0) && (i < FLASH_DEV_BLOCK_SIZE); i++) {
            if (*addr_ptr != flash_xfer_buf[i]) {
                // Only update return value if erase operation was OK
                if (FLASH_ERR_OK == res) res = FLASH_ERR_DRV_VERIFY;
                break;
            }
            addr_ptr++;
            len -= sizeof(*data_ptr);
        }
        // Update data_ptr with the number of bytes written from the caller's
        // buffer (not necessarily the number of bytes written to flash).
        data_ptr = (volatile flash_data_t*)((CYG_ADDRESS)data_ptr + nbytes);
    }

    // Ideally, we'd want to return not only the failure code, but also
    // the address/device that reported the error.
    return res;
}

#endif // CYGONCE_DEVS_FLASH_W29EEXXX_INL
